# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division


class Deserializer(object):

    #! Applicable media types for this deserializer.
    media_types = ()

    def deserialize(self, request=None, text=None):
        """Parses the request content into a format consumable by python.

        @throws ValueError
            To indicate this deserializer cannot deserialize the
            passed text.
        """

        if text is None:
            # Read in the text from the request.
            text = request.read()

        return text



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import six
from importlib import import_module


class Relationship(object):
    """
    Track a relationship on a specific key of a target object.

    NOTE: The idea is to do eventual self-discovery of the resource argument.
    """

    def __init__(self, key, resource, link=True, write=False, list=False):
        self.key = key
        self._resource = resource
        self._resource_cls = None
        self.link = link
        self.list = list
        self.write = write

    @property
    def resource(self):
        if not self._resource_cls:
            if isinstance(self._resource, six.string_types):
                parts = self._resource.split('.')
                mod = import_module('.'.join(parts[:-1]))
                self._resource_cls = getattr(mod, parts[-1])

            else:
                self._resource_cls = self._resource

        return self._resource_cls



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import six
import logging
from collections import Sequence, MutableSequence, Iterable
from armet import http, pagination
from armet.exceptions import ValidationError
from armet.resources.resource import base


logger = logging.getLogger(__name__)


class ManagedResource(base.Resource):
    """Implements the RESTful resource protocol for managed resources.

    @note
        This is not the class to derive from when implementing your own
        resources. Derive from `armet.resources.ManagedResource` (defined in
        the `__init__.py`).
    """

    class Meta:
        # The managed resource access pattern.
        # Traversal is handled a bit more dyanmically than allowed with
        # the simple pattern syntax.
        patterns = [
            r'^'
            r'(?:\:(?P<directives>[^/\(\)]*))?'
            r'(?:\((?P<query>[^/]*)\))?'
            r'(?:/(?P<slug>[^/]+?))?'
            r'(?:/(?P<path>.+?))??'
            r'(?:\.(?P<extensions>[^/]+?))??/??$'
        ]

    #! An ordered dictionary of attributes that are gathered from
    #! attributes defined directly on the resource.
    #!
    #! @code
    #! from armet import resources
    #!
    #! class Resource(resources.Resource):
    #!     name = resources.Attribute()
    #!     created = resources.Attribute('created')
    #! @endcode
    attributes = None

    #! The slug that is used to identify the resource an item is being
    #! requested. None if a list is being requested.
    slug = None

    @classmethod
    def parse(cls, path):
        result = super(ManagedResource, cls).parse(path)
        if result:
            # Found something; parse the result.
            resource, data, end = result

            # Normalize the list arguments.
            for sep, name in (('.', 'extensions'), (':', 'directives')):
                if data[name]:
                    data[name] = data[name].split(sep)

                else:
                    data[name] = []

            # Reset the result.
            result = resource, data, end

        # Return what we got.
        return result

    def __init__(self, *args, **kwargs):
        super(ManagedResource, self).__init__(*args, **kwargs)

        #! Map of errors that have occurred in the clean cycle.
        self._errors = {}

    @property
    def allowed_operations(self):
        """Retrieves the allowed operations for this request."""
        if self.slug is not None:
            return self.meta.detail_allowed_operations

        return self.meta.list_allowed_operations

    def assert_operations(self, *args):
        """Assets if the requested operations are allowed in this context."""
        if not set(args).issubset(self.allowed_operations):
            raise http.exceptions.Forbidden()

    def make_response(self, data=None):
        """Fills the response object from the passed data."""
        if data is not None:
            # Prepare the data for transmission.
            data = self.prepare(data)

            # Encode the data using a desired encoder.
            self.response.write(data, serialize=True)

    def prepare(self, data):
        if data is None:
            # No data; return nothing.
            return None

        if self.path:
            # Ensure we are dealing with only one segment.
            if '/' in self.path:
                raise http.exceptions.NotFound()

        if (isinstance(data, Iterable)
                and not isinstance(data, six.string_types)):
            # Resolve the sequence only if we need to.
            if not isinstance(data, MutableSequence):
                data = list(data)

            # Attempt to prepare each item of the iterable (as long as
            # we're not a string or some sort of mapping).
            for index, value in enumerate(data):
                data[index] = self.item_prepare(data[index])

            return data

        # Prepare just the singular value and return.
        data = self.item_prepare(data)
        if data is None:
            raise http.exceptions.NotFound()

        return data

    def attribute_prepare(self, name, attribute, item):
        # Run the attribute through its prepare cycle.
        return attribute.prepare(
            # Optional preparation cycle on the resource.
            self.preparers[name](
                # Retrieves the value from the object.
                self, item, attribute.get(item)))

    def item_prepare(self, item):
        # If we are the root resource, clear the map.
        if type(self) == self.request._resource.__class__:
            self.request._embed_related.clear()

        # We've started doing this one.
        self.request._embed_related.add(type(self))

        # Check for a path first.
        if self.path:
            attribute = self.attributes.get(self.path)
            if attribute is None:
                raise http.exceptions.NotFound()

            if not attribute.read:
                raise http.exceptions.Forbidden()

            return self.attribute_prepare(self.path, attribute, item)

        # Initialize the object that hold the resultant item.
        obj = {}

        # Iterate through the attributes and build the object from the item.
        for name, attribute in self.attributes.items():
            if attribute.include:
                # Run the attribute through its prepare cycle.
                obj[attribute.name] = self.attribute_prepare(
                    name, attribute, item)

        if type(self) == self.request._resource.__class__:
            # Iterate through the relationships and build their objects.
            # FIXME: This should be opt-in
            for key, relationship in self.relationships.items():
                # If we did this one; skip
                if relationship.resource in self.request._embed_related:
                    continue

                # If we are in list access and this is not okay for
                # lists then skip.
                if self.slug is None and not relationship.list:
                    continue

                # Say that we did this one.
                self.request._embed_related.add(relationship.resource)

                # Construct the related resource
                related = relationship.resource(
                    self.request, self.response)
                # related.require_authentication(self.request)

                # Get the related items.
                related_items = self.read_related(
                    item, related, relationship.key)

                # Prepare and add to the attribute.
                obj[key] = related.prepare(related_items)

                # Say that we're done with this one.
                self.request._embed_related = {
                    x for x in self.request._embed_related
                    if not issubclass(x, relationship.resource)
                }

            # TODO: Remove all that we are.
            self.request._embed_related = {
                x for x in self.request._embed_related
                if not isinstance(self, x)
            }

        # Return the resultant object.
        return obj

    def _clean(self, target, data):
        # Wrap clean so that it can be extended and have validation
        # errors properly handled.

        # HACK: Replace this later with passing item down through clean(...) --
        # however this is to fix a bug and should not break the API.
        self.__target = target

        try:
            data = self.clean(data)

        except AssertionError as ex:
            self._errors['__all__'] = [str(ex)]

        except ValidationError as ex:
            self._errors['__all__'] = ex.errors

        if self._errors:
            # Collect errors and raise a BadRequest message to the client.
            raise http.exceptions.BadRequest(self._errors)

        return data

    def clean(self, data):
        if not data:
            # If no data; just alias it to an empty dictionary.
            data = {}

        if (isinstance(data, Sequence)
                and not isinstance(data, six.string_types)):
            # Attempt to clean each item of the iterable (as long as
            # we're not a string or some sort of mapping).
            for index, value in enumerate(data):
                data[index] = self.item_clean(data[index])

            return data

        # Clean just the singular value and return.
        return self.item_clean(data)

    def item_clean(self, item):
        # Iterate through the attributes and build the object from the item.
        obj = {}

        for name, attribute in self.attributes.items():
            value = item.get(attribute.name)

            try:
                if value is not None:
                    # Run the attribute through its clean cycle.
                    value = self.cleaners[name](
                        # Micro preparation cycle on the attribute object.
                        self, attribute.clean(value))

                    # Check if this attribute is writeable.
                    if not attribute.write:
                        if (not self.__target
                                or attribute.get(self.__target) != value):
                            raise ValidationError('Attribute is read-only.')

                # Ensure that we don't have a null or it is provided.
                if value is None or (
                        isinstance(value, six.string_types) and value == ''):
                    if name in item and not attribute.null:
                        raise ValidationError('Must not be null.')

                    elif attribute.required:
                        raise ValidationError('Must be provided.')

            except (ValueError, AssertionError) as ex:
                self._errors[attribute.name] = [str(ex)]
                value = None

            except ValidationError as ex:
                self._errors[attribute.name] = ex.errors
                value = None

            obj[name] = value

        # Iterate through write-able relations
        for name, relation in six.iteritems(self.relationships):
            if relation.write:
                value = item.get(name)
                try:
                    value = self.clean_related(relation, value)
                    obj[name] = value

                except (ValueError, AssertionError) as ex:
                    self._errors[attribute.name] = [str(ex)]
                    value = None

                except ValidationError as ex:
                    self._errors[attribute.name] = ex.errors
                    value = None

        return obj

    @property
    def http_allowed_methods(self):
        if self.slug is None:
            # No slug means that we're accessing this as a list.
            return self.meta.http_list_allowed_methods

        # A slug exists; this is being accessed as an item.
        return self.meta.http_detail_allowed_methods

    def require_http_allowed_method(self, request):
        # No super call as the following replaces it by checking
        # only against the more specific `list_*` or `detail_*` http
        # allowed methods.

        # Check against `list_*` or `detail_*` allowed methods.
        if request.method not in self.http_allowed_methods:
            # Current method is found to not be allowed for
            # this type of access; raise.
            raise http.exceptions.MethodNotAllowed(self.http_allowed_methods)

    def get(self, request, response):
        """Processes a `GET` request."""
        # Ensure we're allowed to read the resource.
        self.assert_operations('read')

        # Delegate to `read` to retrieve the items.
        items = self.read()

        # if self.slug is not None and not items:
        #     # Requested a specific resource but nothing is returned.

        #     # Attempt to resolve by changing what we understand as
        #     # a slug to a path.
        #     self.path = self.path + self.slug if self.path else self.slug
        #     self.slug = None

        #     # Attempt to retreive the resource again.
        #     items = self.read()

        # Ensure that if we have a slug and still no items that a 404
        # is rasied appropriately.
        if not items:
            raise http.exceptions.NotFound()

        if (isinstance(items, Iterable)
                and not isinstance(items, six.string_types)) and items:
            # Paginate over the collection.
            items = pagination.paginate(self.request, self.response, items)

        # Build the response object.
        self.make_response(items)

    def post(self, request, response):
        """Processes a `POST` request."""
        if self.slug is not None:
            # Don't know what to do an item access.
            raise http.exceptions.NotImplemented()

        # Ensure we're allowed to create a resource.
        self.assert_operations('create')

        # Deserialize and clean the incoming object.
        data = self._clean(None, self.request.read(deserialize=True))

        # Delegate to `create` to create the item.
        item = self.create(data)

        # Build the response object.
        self.response.status = http.client.CREATED
        self.make_response(item)

    def put(self, request, response):
        """Processes a `PUT` request."""
        if self.slug is None:
            # Mass-PUT is not implemented.
            raise http.exceptions.NotImplemented()

        # Check if the resource exists.
        target = self.read()

        # Deserialize and clean the incoming object.
        data = self._clean(target, self.request.read(deserialize=True))

        if target is not None:
            # Ensure we're allowed to update the resource.
            self.assert_operations('update')

            try:
                # Delegate to `update` to create the item.
                self.update(target, data)

            except AttributeError:
                # No read method defined.
                raise http.exceptions.NotImplemented()

            # Build the response object.
            self.make_response(target)

        else:
            # Ensure we're allowed to create the resource.
            self.assert_operations('create')

            # Delegate to `create` to create the item.
            target = self.create(data)

            # Build the response object.
            self.response.status = http.client.CREATED
            self.make_response(target)

    def delete(self, request, response):
        """Processes a `DELETE` request."""
        if self.slug is None:
            # Mass-DELETE is not implemented.
            raise http.exceptions.NotImplemented()

        # Ensure we're allowed to destroy a resource.
        self.assert_operations('destroy')

        # Delegate to `destroy` to destroy the item.
        self.destroy()

        # Build the response object.
        self.response.status = http.client.NO_CONTENT
        self.make_response()

    def _parse_link_headers(self, header):
        # Collect all the passed link headers.
        links = []
        for link_header in header.split(','):
            link = link_header.split(';')
            attrs = {k.strip(): v for (k, v) in [
                     x.split('=') for x in link[1:]]}
            attrs['uri'] = link[0][1:-1]
            links.append(attrs)

        # Return the collected links.
        return links

    def link(self, request, response):
        """Processes a `LINK` request.

        A `LINK` request is asking to create a relation from the currently
        represented URI to all of the `Link` request headers.
        """
        from armet.resources.managed.request import read

        if self.slug is None:
            # Mass-LINK is not implemented.
            raise http.exceptions.NotImplemented()

        # Get the current target.
        target = self.read()

        # Collect all the passed link headers.
        links = self._parse_link_headers(request['Link'])

        # Pull targets for each represented link.
        for link in links:
            # Delegate to a connector.
            self.relate(target, read(self, link['uri']))

        # Build the response object.
        self.response.status = http.client.NO_CONTENT
        self.make_response()

    def unlink(self, request, response):
        """Processes a `UNLINK` request.

        A `UNLINK` request is asking to revoke a relation from the currently
        represented URI to all of the `Link` request headers.
        """
        from armet.resources.managed.request import read

        if self.slug is None:
            # Mass-LINK is not implemented.
            raise http.exceptions.NotImplemented()

        # Get the current target.
        target = self.read()

        # Collect all the passed link headers.
        links = self._parse_link_headers(request['Link'])

        # Pull targets for each represented link.
        for link in links:
            # Delegate to a connector.
            self.unrelate(target, read(self, link['uri']))

        # Build the response object.
        self.response.status = http.client.NO_CONTENT
        self.make_response()



# -*- coding: utf-8 -*-
"""Tuples of the accepted media types for the various supported content types.

The first media type of each respective listing is what armet will
provide for its `Content-Type` upon serialization; armet will accept
any of the listed types when receiving content or when asked to
serialize in a specific format.
"""
from __future__ import absolute_import, unicode_literals, division

#! Multipart form data; as defined by HTML5.
#!
#! http://www.w3.org/TR/html5/constraints.html#multipart-form-data
FORM_DATA = 'multipart/form-data',

#! URL encoded form data; as defined by HTML5.
#!
#! http://www.w3.org/TR/html5/constraints.html#url-encoded-form-data
URL = 'application/x-www-form-urlencoded',

#! JavaScript Object Notation, is a text-based open standard designed
#! for human-readable data interchange.
JSON = (
    # Offical; as per RFC 4627.
    'application/json',

    # Widely used (thanks <http://www.json.org/JSONRequest.html>.)
    'application/jsonrequest',

    # Miscellaneous mimetypes that are used frequently (incorrectly).
    'application/x-json',
    'text/json; charset=utf-8',

    # Widely used (incorrectly) thanks to ruby.
    'text/x-json; charset=utf-8',
)

#! Extensible Markup Language (XML) is a markup language that defines a set
#! of rules for encoding documents in a format that is both human-readable
#! and machine-readable.
XML = (
    # To be used when the XML is NOT human-readable as per RFC 3023.
    # Our XML is not considered human-readable as it is minified before
    # transfer.
    'application/xml',

    # To be used when the XML is human-readable as per RFC 3023.
    'text/xml; charset=utf-8',
)

#! YAML is a human friendly data serialization standard
#! for all programming languages.
#!
#! @note
#!      There is no standard media type for YAML.
YAML = (
    'application/x-yaml',
    'application/yaml',
    'text/yaml; charset=utf-8',
    'text/x-yaml; charset=utf-8',
)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
from armet import utils
from pytest import mark


@mark.bench('utils.cons', number=1000000)
class TestConstruction:

    def test_mapping(self):
        lhs = {'x': 10, 'y': 20}
        rhs = {'z': 15}

        value = utils.cons(lhs, rhs)

        assert value == {'x': 10, 'y': 20, 'z': 15}

    def test_list(self):
        lhs = [10, 20]
        rhs = [15]

        value = utils.cons(lhs, rhs)

        assert value == [10, 20, 15]

    def test_scalar(self):
        lhs = [10, 20]
        rhs = 15

        value = utils.cons(lhs, rhs)

        assert value == [10, 20, 15]

    def test_string(self):
        lhs = ['10', 20]
        rhs = '15'

        value = utils.cons(lhs, rhs)

        assert value == ['10', 20, '15']


@mark.bench('utils.dasherize', number=1000000)
class TestDasherize:

    def test_word(self):
        assert utils.dasherize('word') == 'word'

    def test_camel(self):
        assert utils.dasherize('camelCase') == 'camel-case'

    def test_pascal(self):
        assert utils.dasherize('PascalCase') == 'pascal-case'

    def test_underscore(self):
        assert utils.dasherize('under_score') == 'under-score'

    def test_dashed(self):
        assert utils.dasherize('dashed-words') == 'dashed-words'



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import ujson as json
import six
import uuid
from armet import serializers
from pytest import mark, raises


class TestSerializer:

    media_type = None

    Serializer = None

    @classmethod
    def setup_class(cls):
        cls.serializer = cls.Serializer()

    def serialize(self, data):
        content = self.serializer.serialize(data)
        if content and isinstance(content, six.binary_type):
            content = content.decode('utf8')

        self.content = content


@mark.bench('self.serializer.serialize', iterations=10000)
class TestJSONSerializer(TestSerializer):

    media_type = 'application/json'

    Serializer = serializers.JSONSerializer

    def test_none(self):
        self.serialize(None)

        assert self.content == '{}'

    def test_number(self):
        self.serialize(42)

        assert self.content == '[42]'

    def test_boolean(self):
        self.serialize(True)

        assert self.content == '[true]'

        self.serialize(False)

        assert self.content == '[false]'

    def test_array(self):
        self.serialize([1, 2, 3])

        assert self.content == '[1,2,3]'

    def test_array_nested(self):
        self.serialize([1, [2, 4, 5], 3])

        assert self.content == '[1,[2,4,5],3]'

    def test_dict(self):
        message = {'x': 1, 'y': 2}
        self.serialize(message)

        assert json.loads(self.content) == message

    def test_generator(self):
        self.serialize(x for x in range(10))

        assert self.content == '[0,1,2,3,4,5,6,7,8,9]'

    def test_large(self):
        payload_item = {
            'organization': uuid.uuid4().hex,
            'user': uuid.uuid4().hex,
            'id': uuid.uuid4().hex,
            'is_active': False,
            'is_pending': True,
            'members': [
                {
                    'id': uuid.uuid4().hex,
                    'organization': uuid.uuid4().hex,
                    'is_dead': False
                }
            ],
        }

        payload = [payload_item] * 100

        self.serialize(payload)

        assert self.content == json.dumps(payload, ensure_ascii=False)


@mark.bench('self.serializer.serialize', iterations=10000)
class TestURLSerializer(TestSerializer):

    media_type = 'application/x-www-form-urlencoded'

    Serializer = serializers.URLSerializer

    def test_none(self):
        self.serialize(None)

        assert self.content == ''

    def test_nested(self):
        self.serialize({"foo": [1, 2, 3]})

        assert self.content == 'foo=1&foo=2&foo=3'

    def test_impossible(self):
        data = [{"foo": "bar"}, {"bar": "baz"}]
        with raises(ValueError):
            self.serialize(data)

    def test_tuple(self):
        self.serialize([('foo', 'bar'), ('bar', 'baz')])

        assert self.content == "foo=bar&bar=baz"



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
from django.conf.urls import patterns, include, url
from ..utils import force_import_module


# Initial URL configuration.
urlpatterns = patterns('')

# Import the resources; iterate and mount each one.
module = force_import_module('tests.connectors.resources')
for name in module.__all__:
    cls = getattr(module, name)
    urlpatterns += patterns('', url(r'^api/', include(cls.urls)))



# -*- coding: utf-8 -*-
from .base import Deserializer
from .json import JSONDeserializer
from .url import URLDeserializer

__all__ = [
    'Deserializer',
    'JSONDeserializer',
    'URLDeserializer'
]



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import six
from .base import Resource as BaseResource
from .meta import ResourceBase

__all__ = [
    'Resource'
]


class Resource(six.with_metaclass(ResourceBase, BaseResource)):
    """Implements the RESTful resource protocol for abstract resources.
    """



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
from armet import http
from .base import BaseResourceTest


class TestResourceDelete(BaseResourceTest):

    def test_delete_existing(self, connectors):
        response, content = self.client.delete(
            path='/api/poll/1/',
            headers={'Content-Type': 'application/json'})

        assert response.status == http.client.NO_CONTENT



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division


def read(resource, url):
    """
    Perform a `read` request in the passed `resource` context against
    the given `url`.

    Returns what a `read` would return (the managed target item).
    """
    return resource._request_read(url)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
from armet import http
from .base import BaseResourceTest
import pytest

# Shortcut to the skipif marker.
skipif = pytest.mark.skipif


class TestResourceAccess(BaseResourceTest):

    def test_simple(self, connectors):
        response, _ = self.client.get('/api/simple/')

        assert response.status == http.client.OK

    def test_redirect(self, connectors):
        response, _ = self.client.get('/api/simple')

        assert response.status == http.client.MOVED_PERMANENTLY

        uri = 'http://{}:{}/api/simple/'.format(self.host, self.port)
        assert response.get('location') == uri

    def test_redirect_trailing(self, connectors):
        response, _ = self.client.get('/api/simple-trailing/')

        assert response.status == http.client.MOVED_PERMANENTLY

        uri = 'http://{}:{}/api/simple-trailing'.format(self.host, self.port)
        assert response.get('location') == uri

    # NOTE: The below test fails in flask / werkzeug and I don't want
    #   to add crazy hacks to make it work.
    # <https://github.com/mitsuhiko/werkzeug/issues/402>
    # def test_redirect_complex(self, connectors):
    #     response, _ = self.client.get('/api/simple:hello(x=y)?x=3&y=4')

    #     assert response.status == http.client.MOVED_PERMANENTLY

    #     uri = 'http://{}:{}/api/simple:hello(x=y)/?x=3&y=4'.format(self.host,
    #                                                                self.port)
    #     assert response.get('location') == uri

    def test_not_found(self, connectors):
        response, _ = self.client.get('/api/unknown')

        assert response.status == http.client.NOT_FOUND

        response, _ = self.client.get('/api/unknown/')

        assert response.status == http.client.NOT_FOUND

    def test_unknown(self, connectors):
        response, _ = self.client.request('/api/simple/', method='APPLE')

        assert response.status == http.client.METHOD_NOT_ALLOWED

    def test_not_allowed(self, connectors):
        response, _ = self.client.request('/api/simple/', method='CONNECT')

        assert response.status == http.client.METHOD_NOT_ALLOWED

    def test_not_implemented(self, connectors):
        response, _ = self.client.request('/api/simple/', method='PATCH')

        assert response.status == http.client.NOT_IMPLEMENTED

    def test_streaming(self, connectors):
        response, content = self.client.request('/api/streaming/')
        data = content.decode('utf-8')

        assert response.status == 412
        assert response.get('content-type') == 'text/plain'
        assert data == 'this\nwhere\nwhence\nthat\nwhy\nand the other'

    # @skipif("sys.version_info >= (3, 0)")
    # @skipif("__import__('platform').python_implementation() == 'PyPy'")
    # def test_async(self, connectors):
    #     response, content = self.client.request('/api/async/')
    #     data = content.decode('utf-8')

    #     assert response.status == 412
    #     assert response.get('content-type') == 'text/plain'
    #     assert data == 'Hello'

    # @skipif("sys.version_info >= (3, 0)")
    # @skipif("__import__('platform').python_implementation() == 'PyPy'")
    # def test_async_stream(self, connectors):
    #     response, content = self.client.request('/api/async-stream/')
    #     content = content.decode('utf-8')

    #     assert response.status == 412
    #     assert response.get('content-type') == 'text/plain'
    #     assert content == 'this\nwhere\nwhence\nthat\nwhy\nand the other'


class TestResourceLightweight(BaseResourceTest):

    def test_lightweight(self, connectors):
        response, content = self.client.get('/api/lightweight/')
        data = content.decode('utf-8')

        assert response.status == 412
        assert response.get('content-type') == 'text/plain'
        assert data == 'Hello'

    def test_lightweight_post(self, connectors):
        response, content = self.client.post('/api/lightweight/')
        data = content.decode('utf-8')

        assert response.status == 414
        assert response.get('content-type') == 'text/plain'
        assert data == 'Hello POST'

    def test_lightweight_streaming(self, connectors):
        response, content = self.client.request('/api/lightweight-streaming/')
        data = content.decode('utf-8')

        assert response.status == 412
        assert response.get('content-type') == 'text/plain'
        assert data == 'this\nwhere\nwhence\nthat\nwhy\nand the other'

    # @skipif("sys.version_info >= (3, 0)")
    # @skipif("__import__('platform').python_implementation() == 'PyPy'")
    # def test_lightweight_async(self, connectors):
    #     response, content = self.client.request('/api/lightweight-async/')
    #     data = content.decode('utf-8')

    #     assert response.status == 412
    #     assert response.get('content-type') == 'text/plain'
    #     assert data == 'Hello'


class TestResourceCookie(BaseResourceTest):

    def test_send_and_check(self, connectors):
        response, content = self.client.request(
            path='/api/cookie/',
            headers={'Cookie': 'blue=color'})

        data = content.decode('utf-8')

        assert response.status == 200
        assert response.get('content-type') == 'text/plain'
        assert data == 'color'



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import logging
from ..managed import base


logger = logging.getLogger(__name__)


class ModelResource(base.ManagedResource):
    """Implements the RESTful resource protocol for model-bound resources.

    @note
        This is not the class to derive from when implementing your own
        resources. Derive from `armet.resources.ModelResource` (defined in
        the `__init__.py`).
    """



# -*- coding: utf-8 -*-
"""
Describes the authentication protocols and generalizations used to
authenticate access to a resource endpoint.
"""
from __future__ import absolute_import, unicode_literals, division
from armet import http
import base64


class Authentication(object):

    #! Whether returning `None` from `authenticate` indicates
    #! unauthenticated; or, simply no authenticated user.
    allow_anonymous = True

    def __init__(self, **kwargs):
        """Initialize authentication protocol; establish parameters.
        """

        # Allow overriding class attributes.
        self.__dict__.update(kwargs)

    def authenticate(self, request):
        """Gets the a user if they are authenticated; else None.

        @retval False    Unable to authenticate.
        @retval None     Able to authenticate but failed.
        @retval <user>   User object representing the current user.
        """
        return None

    def unauthenticated(self):
        """
        Callback that is invoked when after a user is determined to
        be unauthenticated.
        """
        raise http.exceptions.Forbidden()


class HeaderAuthentication(Authentication):

    def can_authenticate(self, request, method):
        # Determine if we can authenticate.
        return False

    def get_credentials(self, text):
        # Decode credentials.
        return text,

    def get_user(self, request, method, *args):
        """
        Callback that is invoked when a user is attempting to be
        authenticated with a set of credentials.
        """
        return None

    def authenticate(self, request):
        # Retrieve the authorization header.
        header = request.headers.get('Authorization')

        try:
            # Split the authorization header into method and credentials.
            method, text = (header or '').split(' ', 1)

        except ValueError:
            # Strange format in the header.
            return False

        if not self.can_authenticate(request, method):
            # Not the right kind of authentication.
            return False

        # Retreive and return the user object.
        return self.get_user(request, method, *self.get_credentials(text))


class BasicAuthentication(HeaderAuthentication):

    allow_anonymous = False

    def can_authenticate(self, request, method):
        # Determine if we can authenticate.
        return method.lower() == 'basic'

    def get_credentials(self, text):
        try:
            # Decode credentials.
            text = base64.b64decode(text.encode('utf8')).decode('utf8')
            return text.split(':', 1)

        except base64.binascii.Error:
            # Decoding error.
            return None, None



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import six
from six.moves import cStringIO as StringIO
import operator
import itertools
from . import constants


#! Operation to combinator map.
COMBINATORS = {
    constants.LOGICAL_AND: operator.and_,
    constants.LOGICAL_OR: operator.or_}


#! Set of characters that begin an operator.
OPERATOR_BEGIN_CHARS = set(x[0] for _, x in constants.OPERATORS if x)
OPERATOR_BEGIN_CHARS.add(constants.NEGATION[1])

#! Dictionary of operator symbols to operators.
OPERATOR_SYMBOL_MAP = dict((v, k) for k, v in constants.OPERATORS if v)


#! List of operator keywords.
OPERATOR_KEYWORDS = set(k for k, _ in constants.OPERATORS)


# Make a reversed suffix map for convenient stringification.
REVERSED_OPERATOR_SUFFIX_MAP = dict(
    (v, k) for k, v in constants.OPERATOR_SUFFIX_MAP.items()
)


class Query(object):
    """Represents a complete query expression.
    """

    def __init__(self, original, parsed):
        # Keep a copy of the original querystring.
        self.original = original
        self.parsed = parsed


class QuerySegment(object):
    """
    Represents a single query segment with a subject path (`x.a.g`),
    an operator (`=` or `<=`), optional directives (`sort`), and
    a set of values (`5,12,56`).
    """

    def __init__(self, **kwargs):
        #! Path to the attribute being tested (as a list of segments).
        self.path = kwargs.get('path', [])

        #! This is the operator that is being applied to the attribute path.
        self.operator = kwargs.get(
            'operator',
            constants.OPERATOR_MAP[constants.OPERATOR_IEQUAL]
        )

        #! Negation; if this operation has been negated.
        self.negated = kwargs.get('negated', False)

        #! Directives. Directives are a key-value way of specifying commands
        #! on an attribute path.
        self.directives = kwargs.get('directives', [])

        #! Values. Set of values that the attribute path is being checked
        #! against. Only one has to match.
        self.values = kwargs.get('values', [])

    def __repr__(self):
        return str(self)

    def __str__(self):
        """
        Format this query segment in a human-readable representation
        intended for debugging.
        """

        o = StringIO()

        o.write('(')

        if self.negated:
            o.write('not ')

        o.write('.'.join(self.path))

        if self.values:
            o.write(' :%s ' % REVERSED_OPERATOR_SUFFIX_MAP[self.operator])

        o.write(' | '.join(map(lambda x: "'{}'".format(str(x)), self.values)))

        o.write(')')
        return o.getvalue()


class NoopQuerySegment(object):
    """A query segment that doesn't perform an operation.  For the purposes
    of binary and unary combinations, this should be treated as True.
    A NoopQuerySegment should only be encountered when the entire query string
    is missing."""

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "TRUE"


class BinarySegmentCombinator(object):
    """
    Represents the combination of 2 query segments `(x=y)&(y=z)`
    """

    def __init__(self, left, right, operation=operator.and_):
        self.left = left
        self.right = right
        self.operation = operation

    def __repr__(self):
        return str(self)

    def __str__(self):
        combinators = {
            operator.and_: 'AND',
            operator.or_: 'OR'
        }

        return "{} {} {}".format(
            str(self.left),
            combinators[self.operation],
            str(self.right)
        )


class UnarySegmentCombinator(object):
    """
    Represents a unary combination of 2 query segments. `!(x=y)`
    """
    def __init__(self, operand, operation=operator.not_):
        self.operand = operand
        self.operation = operation

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "{} {}".format('NOT', self.operand)


def parse(text, encoding='utf8'):
    """Parse the querystring into a normalized form."""

    # Decode the text if we got bytes.
    if isinstance(text, six.binary_type):
        text = text.decode(encoding)

    return Query(text, split_segments(text))


def reset_stringio(buf):
    # This combination of functions happens way too often in here.
    buf.truncate(0)
    buf.seek(0)


def split_segments(text, closing_paren=False):
    """Return objects representing segments."""
    buf = StringIO()

    # The segments we're building, and the combinators used to combine them.
    # Note that after this is complete, this should be true:
    # len(segments) == len(combinators) + 1
    # Thus we can understand the relationship between segments and combinators
    # like so:
    #  s1 (c1) s2 (c2) s3 (c3) where sN are segments and cN are combination
    # functions.
    # TODO: Figure out exactly where the querystring died and post cool
    # error messages about it.
    segments = []
    combinators = []

    # A flag dictating if the last character we processed was a group.
    # This is used to determine if the next character (being a combinator)
    # is allowed to
    last_group = False

    # The recursive nature of this function relies on keeping track of the
    # state of iteration.  This iterator will be passed down to recursed calls.
    iterator = iter(text)

    # Detection for exclamation points.  only matters for this situation:
    # foo=bar&!(bar=baz)
    last_negation = False

    for character in iterator:
        if character in COMBINATORS:

            if last_negation:
                buf.write(constants.OPERATOR_NEGATION)

            # The string representation of our segment.
            val = buf.getvalue()
            reset_stringio(buf)

            if not last_group and not len(val):
                raise ValueError('Unexpected %s.' % character)

            # When a group happens, the previous value is empty.
            if len(val):
                segments.append(parse_segment(val))

            combinators.append(COMBINATORS[character])

        elif character == constants.GROUP_BEGIN:
            # Recursively go into the next group.

            if buf.tell():
                raise ValueError('Unexpected %s' % character)

            seg = split_segments(iterator, True)

            if last_negation:
                seg = UnarySegmentCombinator(seg)

            segments.append(seg)

            # Flag that the last entry was a grouping, so that we don't panic
            # when the next character is a logical combinator
            last_group = True
            continue

        elif character == constants.GROUP_END:
            # Build the segment for anything remaining, and then combine
            # all the segments.
            val = buf.getvalue()

            # Check for unbalanced parens or an empty thing: foo=bar&();bar=baz
            if not buf.tell() or not closing_paren:
                raise ValueError('Unexpected %s' % character)

            segments.append(parse_segment(val))
            return combine(segments, combinators)

        elif character == constants.OPERATOR_NEGATION and not buf.tell():
            last_negation = True
            continue

        else:
            if last_negation:
                buf.write(constants.OPERATOR_NEGATION)
            if last_group:
                raise ValueError('Unexpected %s' % character)
            buf.write(character)

        last_negation = False
        last_group = False
    else:
        # Check and see if the iterator exited early (unbalanced parens)
        if closing_paren:
            raise ValueError('Expected %s.' % constants.GROUP_END)

        if not last_group:
            # Add the final segment.
            segments.append(parse_segment(buf.getvalue()))

    # Everything completed normally, combine all the segments into one
    # and return them.
    return combine(segments, combinators)


def combine(segments, combinators):
    # We get [a,b,c] in segments and combinators that should be applied as
    # [a(op)b, result(op)c]
    operands = iter(segments)
    operators = iter(combinators)
    first = next(operands)
    reducer = lambda x, y: BinarySegmentCombinator(x, y, next(operators))
    return six.moves.reduce(reducer, operands, first)


def parse_directive(key):
    """
    Takes a key of type (foo:bar) and returns either the key and the
    directive, or the key and None (for no directive.)
    """
    if constants.DIRECTIVE in key:
        return key.split(constants.DIRECTIVE, 1)
    else:
        return key, None


def parse_segment(text):
    "we expect foo=bar"

    if not len(text):
        return NoopQuerySegment()

    q = QuerySegment()

    # First we need to split the segment into key/value pairs.  This is done
    # by attempting to split the sequence for each equality comparison.  Then
    # discard any that did not split properly.  Then chose the smallest key
    # (greedily chose the first comparator we encounter in the string)
    # followed by the smallest value (greedily chose the largest comparator
    # possible.)

    try:
        # import ipdb; ipdb.set_trace()
        # translate into [('=', 'foo=bar')]
        equalities = zip(constants.OPERATOR_EQUALITIES, itertools.repeat(text))
        # Translate into [('=', ['foo', 'bar'])]
        equalities = map(lambda x: (x[0], x[1].split(x[0], 1)), equalities)
        # Remove unsplit entries and translate into [('=': ['foo', 'bar'])]
        # Note that the result from this stage is iterated over twice.
        equalities = list(filter(lambda x: len(x[1]) > 1, equalities))
        # Get the smallest key and use the length of that to remove other items
        key_len = len(min((x[1][0] for x in equalities), key=len))
        equalities = filter(lambda x: len(x[1][0]) == key_len, equalities)

        # Get the smallest value length. thus we have the earliest key and the
        # smallest value.
        op, (key, value) = min(equalities, key=lambda x: len(x[1][1]))
    except ValueError:
        # Only the key exists.  See if there's a directive.
        key, directive = parse_directive(text)
        value = ''
        op = constants.OPERATOR_EQUALITY_FALLBACK
        if directive is None:
            # import ipdb; ipdb.set_trace()
            raise ValueError(
                'Encountered a segment with neither directive nor equality.')

        q.directive = directive

    # Process negation.  This comes in both foo.not= and foo!= forms.
    path = key.split(constants.SEP_PATH)
    last = path[-1]

    # Check for !=
    if last.endswith(constants.OPERATOR_NEGATION):
        last = last[:-1]
        q.negated = not q.negated

    # Check for foo.not=
    if last == constants.PATH_NEGATION:
        path.pop(-1)
        q.negated = not q.negated

    q.values = value.split(constants.SEP_VALUE)

    # Check for suffixed operators (foo.gte=bar).  Prioritize suffixed
    # entries over actual equality checks.
    if path[-1] in constants.OPERATOR_SUFFIXES:

        # The case where foo.gte<=bar, which obviously makes no sense.
        if op not in constants.OPERATOR_FALLBACK:
            raise ValueError(
                'Both path-style operator and equality style operator '
                'provided.  Please provide only a single style operator.')

        q.operator = constants.OPERATOR_SUFFIX_MAP[path[-1]]
        path.pop(-1)
    else:
        q.operator = constants.OPERATOR_EQUALITY_MAP[op]

    if not len(path):
        raise ValueError('No attribute navigation path provided.')

    q.path = path

    return q



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import bottle
import wsgi_intercept
from wsgi_intercept.httplib2_intercept import install
from .utils import force_import_module


def http_setup(connectors, host, port, callback):
    # Install the WSGI interception layer.
    install()

    # Bottle is pretty straightforward.
    # We just need to push an application context.
    application = bottle.Bottle(__name__)
    application.catchall = False

    # Ensure we're debugging.
    bottle.debug(True)

    # Invoke the callback if we got one.
    if callback:
        callback()

    # Then import the resources; iterate and mount each one.
    module = force_import_module('tests.connectors.resources')
    for name in module.__all__:
        getattr(module, name).mount(r'/api/', application)

    # Enable the WSGI interception layer.
    wsgi_intercept.add_wsgi_intercept(host, port, lambda: application)


def http_teardown(host, port):
    # Remove the WSGI interception layer.
    wsgi_intercept.remove_wsgi_intercept(host, port)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import six
from .attribute import Attribute


class BooleanAttribute(Attribute):

    type = bool

    #! Textual values accepted for `True`.
    TRUE = (
        'true',
        't',
        'yes',
        'y',
        'on',
        '1'
    )

    #! Textual values accepted for `False`.
    FALSE = (
        'false',
        'f',
        'no',
        'n',
        'off',
        '0'
    )

    def clean(self, value):
        if value is None:
            # Value is nothing; return it.
            return value

        if value is True or value is False:
            # Value is a python boolean; just return it.
            return value

        if value.strip().lower() in self.TRUE:
            # Some sort of truthy value.
            return True

        if value.strip().lower() in self.FALSE:
            # Some sort of falsy value.
            return False

        # Neither true or false matches; return a boolifyed version of
        # whatever we have.
        return bool(value)


class TextAttribute(Attribute):

    type = six.text_type

    def clean(self, value):
        return six.text_type(value) if value is not None else None


class IntegerAttribute(Attribute):

    type = int

    def clean(self, value):
        if isinstance(value, six.integer_types):
            # Value is already an integer.
            return value

        if isinstance(value, six.string_types):
            # Strip the string of whitespace
            value = value.strip()

        if not value:
            # Value is nothing; return nothing.
            return None

        try:
            # Attempt to coerce whatever we have as an int.
            return int(value)

        except ValueError:
            # Failed to do so.
            raise ValueError('Not a valid integral value.')



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import six
import operator
from functools import partial
from collections import Iterable
from copy import deepcopy
from six.moves import map, reduce
from armet.exceptions import ImproperlyConfigured
from armet.query import parser, Query, QuerySegment, constants
from armet.http.exceptions import BadRequest
from sqlalchemy.exc import InvalidRequestError
import sqlalchemy as sa
import functools


class ModelResourceOptions(object):

    def __init__(self, meta, name, bases):
        #! SQLAlchemy session used to perform operations on the models.
        self.Session = meta.get('Session')
        if not self.Session:
            raise ImproperlyConfigured(
                'A session factory (via sessionmaker) is required by '
                'the SQLAlchemy model connector.')


def ilike_helper(default):
    """Helper function that performs an `ilike` query if a string value
    is passed, otherwise the normal default operation."""
    @functools.wraps(default)
    def wrapped(x, y):
        # String values should use ILIKE queries.
        if isinstance(y, six.string_types):
            return x.ilike("%" + y + "%")
        else:
            return default(x, y)
    return wrapped


# Build an operator map to use for sqlalchemy.
OPERATOR_MAP = {
    constants.OPERATOR_EQUAL: operator.eq,
    constants.OPERATOR_IEQUAL: ilike_helper(operator.eq),
    constants.OPERATOR_LT: operator.lt,
    constants.OPERATOR_GT: operator.gt,
    constants.OPERATOR_LTE: operator.le,
    constants.OPERATOR_GTE: operator.ge,
    constants.OPERATOR_ICONTAINS: ilike_helper(operator.contains),
}

# Rewire the map.
OPERATOR_MAP = dict(
    (constants.OPERATOR_MAP[k], v) for k, v in OPERATOR_MAP.items()
)


def build_segment(model, segment, attr, clean):
    # Get the associated column for the initial path.
    path = segment.path.pop(0)
    col = getattr(model, path)

    # Resolve the inner-most path segment.
    if segment.path:
        if col.impl.accepts_scalar_loader:
            return col.has(build_segment(
                col.property.mapper.class_, segment, attr, clean))

        else:
            try:
                return col.any(build_segment(
                    col.property.mapper.class_, deepcopy(segment),
                    attr, clean))

            except InvalidRequestError:
                return col.has(build_segment(
                    col.property.mapper.class_, deepcopy(segment),
                    attr, clean))

    # Determine the operator.
    op = OPERATOR_MAP[segment.operator]

    # Apply the operator to the values and return the expression
    qs = reduce(operator.or_,
                map(partial(op, col),
                    map(lambda x: clean(attr.try_clean(x)), segment.values)))

    # Apply the negation.
    if segment.negated:
        qs = ~qs

    # Return our query object.
    return qs


def segment_query(resource, segment, attributes, cleaners, model):

    attribute = attributes[segment.path[0]]

    # Modify the segment's to reflect what the attribute was declared for.
    segment.path[0:1] = attribute.path.split('.')

    # Create a cleaner that can work here.
    clean = partial(cleaners[attribute.name], resource)

    # Dispatch to the recursive segment building function
    return build_segment(model, segment, attribute, clean)


def noop_query(*args):
    return sa.sql.true()


def unary_query(resource, query, *args):
    return query.operation(build_clause(resource, query.operand, *args))


def binary_query(resource, query, *args):
    return query.operation(
        build_clause(resource, query.left, *args),
        build_clause(resource, query.right, *args))


CLAUSE_MAP = {
    parser.NoopQuerySegment: noop_query,
    parser.BinarySegmentCombinator: binary_query,
    parser.UnarySegmentCombinator: unary_query,
    parser.QuerySegment: segment_query,
}


def build_clause(resource, query, attributes, cleaners, model):
    class_ = type(query) if not isinstance(query, type) else query
    fn = CLAUSE_MAP.get(class_)
    if fn is not None:
        return fn(resource, query, attributes, cleaners, model)
    elif issubclass(class_, Query):
        return build_clause(
            resource,
            query.parsed,
            attributes,
            cleaners,
            model)
    else:
        raise ValueError('Unable to translate query node %s' % str(query))


class ModelResource(object):
    """Specializes the RESTFul model resource protocol for SQLAlchemy.

    @note
        This is not what you derive from to create resources. Import
        ModelResource from `armet.resources` and derive from that.
    """

    def route(self, *args, **kwargs):
        # Establish a session.
        self.session = session = self.meta.Session()

        try:
            # Continue on with the cycle.
            result = super(ModelResource, self).route(*args, **kwargs)

            # Commit the session.
            session.commit()

            # Return the result.
            return result

        except:
            # Something occurred; rollback the session.
            session.rollback()

            # Re-raise the exception.
            raise

        finally:
            # Close the session.
            session.close()

    def filter(self, clause, queryset):
        # Filter the queryset by the passed clause.
        return queryset.filter(clause).distinct()

    def count(self, queryset):
        # Return the count of the queryset.
        return queryset.count()

    def read(self):
        # Initialize the query to the model.
        queryset = self.session.query(self.meta.model)

        query = None
        if self.slug is not None:
            # This is an item-access (eg. GET //:slug); ignore the
            # query string and generate a query-object based on the slug.
            query = Query(
                original=None,
                parsed=QuerySegment(
                    path=self.meta.slug.path.split('.'),
                    operator=constants.OPERATOR_MAP[constants.OPERATOR_EQUAL],
                    values=[self.slug]
                )
            )

        elif self.request.query:
            # This is a list-access; use the query string and construct
            # a query object from it.
            query = parser.parse(self.request.query)

        # Determine if we need to filter the queryset in some way; and if so,
        # filter it.
        clause = None
        if query is not None:
            clause = build_clause(
                self, query, self.attributes,
                self.cleaners, self.meta.model)

            queryset = self.filter(clause, queryset)

        if self.slug is None:
            # Filter the queryset by asserting authorization.
            queryset = self.meta.authorization.filter(
                self.request.user, 'read', self, queryset)

            # Return the queryset.
            return queryset

        else:
            # Get the item in question.
            item = queryset.first()

            # Sanity check to make sure some item was found.
            if item is None:
                return None

            # Ensure the user is authorized to perform this action.
            authz = self.meta.authorization
            if not authz.is_authorized(self.request.user, 'read', self, item):
                authz.unauthorized()

            # We're good, return the item.
            return item

    def create(self, data):
        # Instantiate a new target.
        target = self.meta.model()

        # Iterate through all attributes and set each one.
        for name, attribute in six.iteritems(self.attributes):
            # Set each one on the target.
            value = data.get(name)
            if value is not None:
                attribute.set(target, value)

        # Iterate through all write-able relations and set each one.
        for name, relation in six.iteritems(self.relationships):
            if relation.write:
                # Set each one on the target.
                value = data.get(name)
                if value is not None:
                    # FIXME: Use some deferred thing that is not `setattr`.
                    setattr(target, relation.key, value)

        # Add the target to the session.
        self.session.add(target)
        self.session.flush()

        # Refresh the target object to avoid inconsistencies with storage.
        self.session.expire(target)

        # Ensure the user is authorized to perform this action.
        authz = self.meta.authorization
        if not authz.is_authorized(self.request.user, 'create', self, target):
            authz.unauthorized()

        # Return the target.
        return target

    def update(self, target, data):
        # Iterate through all attributes and set each one.
        for name, attribute in six.iteritems(self.attributes):
            # Set each one on the target.
            attribute.set(target, data.get(name))

        # Iterate through all write-able relations and set each one.
        for name, relation in six.iteritems(self.relationships):
            if relation.write:
                # Set each one on the target.
                # FIXME: Use some deferred thing that is not `setattr`.
                setattr(target, relation.key, data.get(name))

        # Flush the target and expire attributes.
        self.session.flush()

        # Refresh the target object to avoid inconsistencies with storage.
        self.session.expire(target)

        # Ensure the user is authorized to perform this action.
        authz = self.meta.authorization
        if not authz.is_authorized(self.request.user, 'update', self, target):
            authz.unauthorized()

    def destroy(self):
        # Grab the existing target.
        target = self.read()

        # Ensure the user is authorized to perform this action.
        authz = self.meta.authorization
        if not authz.is_authorized(self.request.user, 'destroy', self, target):
            authz.unauthorized()

        # Remove the object from the session.
        self.session.delete(target)

    def _resolve_relation(self, target):
        # Resolve the relationship key.
        key = None
        related = type(self)._related_models
        res_clss = type(target),
        while res_clss:
            res_cls_next = []
            for res_cls in res_clss:
                if res_cls in related:
                    key = self.relationships[related[res_cls]].key
                    break

                if res_cls.__bases__:
                    res_cls_next.extend(res_cls.__bases__)

            else:
                # Nope; continue
                res_clss = res_cls_next
                continue

            # Able to link
            break

        else:
            # Not able to link
            return None

        # Able to link
        return key

    def relate(self, target, other):
        # Resolve the relationship key.
        key = self._resolve_relation(other)
        if not key:
            raise BadRequest({
                '__all__': "Unable to link a '%s' to a '%s'." % (
                    type(target).__name__, type(other).__name__)})

        # Grab the set_ in question.
        set_ = getattr(target, key)

        # Append the relationship.
        set_.append(other)

        # Flush the target and expire attributes.
        self.session.flush()

        # Refresh the target object to avoid inconsistencies with storage.
        self.session.expire(target)

        # Ensure the user is authorized to perform this action.
        authz = self.meta.authorization
        if not authz.is_authorized(self.request.user, 'update', self, target):
            authz.unauthorized()

    def unrelate(self, target, other):
        # Resolve the relationship key.
        key = self._resolve_relation(other)
        if not key:
            raise BadRequest({
                '__all__': "Unable to link a '%s' to a '%s'." % (
                    type(target).__name__, type(other).__name__)})

        # Grab the set_ in question.
        set_ = getattr(target, key)

        # Append the relationship.
        set_.remove(other)

        # Flush the target and expire attributes.
        self.session.flush()

        # Refresh the target object to avoid inconsistencies with storage.
        self.session.expire(target)

        # Ensure the user is authorized to perform this action.
        authz = self.meta.authorization
        if not authz.is_authorized(self.request.user, 'update', self, target):
            authz.unauthorized()

    def read_related(self, target, resource, key):
        # Grab the set_ in question.
        set_ = getattr(target, key)

        # Build the related items.
        qs = set_

        if isinstance(qs, Iterable):
            if self.request.user:
                # Filter the queryset by asserting authorization.
                qs = resource.meta.authorization.filter(
                    self.request.user, 'read', resource, qs)

        else:
            if self.request.user:
                # Ensure we can access this.
                authz = self.meta.authorization
                if not authz.is_authorized(
                        self.request.user, 'read', resource, qs):
                    authz.unauthorized()

        # Return the queryset.
        return qs

    def clean_related(self, relation, value):
        # Grab the model in question.
        model = relation.resource.meta.model

        # Attempt to build a query to `get` the model.
        target = self.session.query(model).get(value)
        if target is not None:
            # Found a target from the value.
            return target

        # Found nothing.



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
from ..managed import meta
from . import options
import six


# Mapping of the canonical resource for a model.
_canonical_resources = {}


class ModelResourceBase(meta.ManagedResourceBase):

    options = options.ModelResourceOptions

    connectors = ['http', 'model']

    _related_models_cache = None

    def __new__(cls, name, bases, attrs):
        # Construct the class object.
        self = super(ModelResourceBase, cls).__new__(cls, name, bases, attrs)

        if self.meta and not self.meta.abstract:
            # Add this to the canonical resource dictionary.
            _canonical_resources[self.meta.model] = self

        # Return the constructed class object.
        return self

    @property
    def _related_models(self):
        if not self._related_models_cache:
            if hasattr(self, 'relationships'):
                # Construct a mapping of all relationships and their models.
                self._related_models_cache = {}
                for key, relation in six.iteritems(self.relationships):
                    self._related_models_cache[
                        relation.resource.meta.model] = key

        return self._related_models_cache



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
from armet.authorization import ManagedAuthorization
from armet.exceptions import ImproperlyConfigured


try:
    import shield

except ImportError:
    shield = None


class ShieldAuthorization(ManagedAuthorization):
    """Implements managed authorization using the shield permissions framework.
    """

    def __init__(self, permissions=None):
        # Ensure we have access to the shield framework.
        if shield is None:
            raise ImproperlyConfigured(
                "'shield' is required to use 'ShieldAuthorization'")

        #! Permissions to check for each operation.
        self.permissions = permissions
        if self.permissions is None:
            self.permissions = {
                'read': ('read',),
                'create': ('create',),
                'update': ('update',),
                'destroy': ('destroy',)
            }

    def is_authorized(self, user, operation, resource, item):
        return shield.has(
            *self.permissions[operation],
            bearer=user,
            target=item)

    def filter(self, user, operation, resource, iterable):
        query = shield.filter(
            *self.permissions[operation],
            bearer=user,
            target=resource.meta.model,
            query=iterable)

        return query.distinct()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import os
import json
import armet
import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.declarative import declarative_base

# Instantiate the declarative base.
Base = declarative_base()

# Instantiate the engine used to access the models.
engine = sa.create_engine('sqlite:///:memory:')

# Construct the session factory.
Session = orm.sessionmaker(bind=engine)


class Poll(Base):

    __tablename__ = 'poll'

    id = sa.Column(sa.Integer, primary_key=True)

    question = sa.Column(sa.String(1024))

    available = sa.Column(sa.Boolean)

    votes = sa.Column(sa.Integer)


def _load_fixture(filename):
    """
    Loads the passed fixture into the database following the
    django format.
    """

    # Read the binary data into text
    with open(filename, 'rb') as stream:
        content = stream.read().decode('utf-8')

    # Decode the data as JSON
    data = json.loads(content)

    # Instantiate a session.
    session = Session()

    # Iterate through the entries to add them one by one.
    for item in data:
        # Resolve model from the table reference.
        table = Base.metadata.tables[item['model'].split('.')[-1]]

        # Add the primary key.
        item['fields']['id'] = item['pk']

        # Add a new row.
        session.connection().execute(table.insert().values(**item['fields']))

    # Commit the session to the database.
    session.commit()


def model_setup():
    # Initialize the database and create all models.
    Base.metadata.drop_all(engine)
    Base.metadata.create_all(engine)

    # Load the data fixture.
    _load_fixture(os.path.join(os.path.dirname(__file__), 'data.json'))

    # Configure armet and provide the session factory.
    armet.use(Session=Session)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import bottle
from . import http


class Resource(object):

    @classmethod
    def view(cls, *args, **kwargs):
        # Construct request and response wrappers.
        async = cls.meta.asynchronous
        request = http.Request(path=kwargs.get('path', ''), asynchronous=async)
        response = http.Response(asynchronous=async)

        # Defer the execution thread if we're running asynchronously.
        if async:
            # Defer the view to pass of control.
            import gevent
            gevent.spawn(super(Resource, cls).view, request, response)

            # Construct and return a streamer.
            return cls.stream(response, response)

        # Pass control off to the resource handler.
        return super(Resource, cls).view(request, response)

    @classmethod
    def mount(cls, url='/', application=None):
        if application is None:
            # If no explicit application is passed; use
            # the current default application.
            application = bottle.app[-1]

        # Generate a name to use to mount this resource.
        name = '{}:{}:{}'.format('armet', cls.__module__, cls.meta.name)

        # Apply the routing rules and add the URL route.
        rule = '{}{}<path:re:$|(?:[/:(.].*)>'.format(url, cls.meta.name)
        application.route(rule, 'ANY', cls.view, name)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division


class classproperty(object):
    """Declares a read-only `property` that acts on the class object.
    """

    def __init__(self, getter):
        self.getter = getter

    def __get__(self, obj, cls):
        return self.getter(cls)


class boundmethod(object):
    """
    Declares a method that can be invoked as both an instance method
    and a class method.
    """

    def __init__(self, method):
        self.method = method

    def __get__(self, obj, cls):
        self.obj = obj if obj else cls
        return self

    def __call__(self, *args, **kwargs):
        return self.method(self.obj, *args, **kwargs)
